##
# c.loader - a loader for automatically compiling C extensions
##
import os.path
import sys
import imp
import subprocess
import sysconfig

class CompilationError(ImportError):
	def __init__(self,
		module_path, source_path,
		compile_log, return_code,
	):
		self.module_path = module_path
		self.source_path = source_path
		self.compile_log = compile_log
		self.return_code = return_code

	def __str__(self):
		msg = "could not compile '{0.module_path}' from source '{0.source_path}'\n".format(self)
		msg += "*** compiler exited with code '{0.return_code}' and emitted the following:\n  ".format(self)
		msg += '  '.join(self.compile_log.splitlines(True))
		msg += '\n***'
		return msg

class CLoader(object):
	"""
	Compile C-API modules into shared libraries.
	"""
	target = sys.platform

	types = [
		('c', '.py.c'),
		('objc', '.py.m'),
		('c++', '.py.c++'),
	]

	@staticmethod
	def build_call(type, target, filename):
		"""
		Construct the parameters to be used to compile and link the new executable.
		"""
		if sys.platform == 'darwin' and type == 'objc':
			# assume macosx
			frameworks = ('-framework', 'CoreFoundation', '-framework', 'Foundation')
		else:
			frameworks = ()
		libs = tuple((sysconfig.get_config_var('SHLIBS') or '').split())
		syslibs = tuple((sysconfig.get_config_var('SYSLIBS') or '').split())
		ldflags = tuple((sysconfig.get_config_var('LDFLAGS') or '').split())
		pyversion = sysconfig.get_config_var('VERSION') or ''.join(map(str, sys.version_info[:2]))
		pyabi = sysconfig.get_config_var('ABIFLAGS') or ''
		pyspec = 'python' + pyversion + pyabi
		return (
			sysconfig.get_config_var('CC'),
			'-v', '-bundle', '-g',
			'-o', target,
		) + libs + syslibs + ldflags + frameworks + (
			'-I' + sysconfig.get_config_var('INCLUDEPY'),
			'-L' + sysconfig.get_config_var('LIBDIR'),
			'-l' + pyspec,
			filename,
		)

	def compile(self):
		compiler = self.build_call(self.type, self.dll, self.source)
		logfile = self.logfile
		cmdfile = self.cmdfile
		if os.path.exists(logfile):
			os.unlink(logfile)
		if os.path.exists(cmdfile):
			os.unlink(cmdfile)
		with open(self.logfile, 'wb') as log:
			compiling = subprocess.Popen(compiler, stdout = log, stderr = subprocess.STDOUT)
			# note command being ran
			with open(self.cmdfile, 'w') as cmdlog:
				cmdlog.write(' '.join(compiler))
			return compiling.wait()

	@property
	def logfile(self):
		'file recording the output of the command that built the extension module'
		return self.dll + ':log'

	@property
	def cmdfile(self):
		'file recording the command issued to build the extension module'
		return self.dll + ':cmd'

	def __init__(self, pkg, name, source, type = 'c'):
		self.type = type
		self.package = pkg
		self.name = name
		self.fullname = '.'.join((pkg,name))
		self.source = source
		if hasattr(imp, 'cache_from_source'):
			self.dll = imp.cache_from_source(self.source)[:-len('.pyc')]
		else:
			self.dll = self.source[:-len('.py.c')]
		dir = os.path.dirname(self.dll)
		if not os.path.exists(dir):
			os.mkdir(dir)
		# XXX: need to manage local ABI/env flags as well
		self.dll += (sysconfig.get_config_var('ABIFLAGS') or '') + '.so'

	@classmethod
	def find_module(typ, fullname, paths = None,
		_realpath = os.path.realpath,
		_join = os.path.join,
		_exists = os.path.exists,
	):
		if paths is None:
			return
		names = fullname.split('.')
		modname = names[-1]
		for f in paths:
			path = _join(_realpath(f), modname)
			for type, ext in typ.types:
				source = path + ext
				if _exists(source):
					return typ('.'.join(names[:-1]), modname, source, type = type)

	def get_code(self, module):
		pass

	def get_data(self, module):
		pass

	def get_filename(self, name):
		return self.source

	def get_source(self, name):
		pass

	def is_package(self, *args, **kw):
		'cloader does not handle packages'
		return False

	def compile_module(self):
		rc = self.compile()
		if rc != 0 or not os.path.exists(self.dll):
			with open(self.logfile, 'rb') as log:
				output = log.read().decode('utf-8', errors='replace').strip()
			raise CompilationError(self.fullname, self.source, output, rc)

	def load_module(self, fullname):
		# instance exists, so compilation succeeded
		if fullname in sys.modules:
			return sys.modules[fullname]
		if not os.path.exists(self.dll) or \
		os.path.getmtime(self.dll) < os.path.getmtime(self.source):
			self.compile_module()
		# compiled "successfully" or already compiled.
		mod = sys.modules[fullname] = imp.load_dynamic(self.name, self.dll)
		mod.__cached__ = self.dll
		mod.__loader__ = self
		return mod
